@@ -39,14 +39,12 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 """
 
-__version__ = "0.0.0"
+__version__ = "0.1.0"
 
-import base64
 from rdflib.term import Identifier as ID
 from rdflib import URIRef as URI
-from rdflib import BNode
-from rdflib import Literal
-from rdflib import RDF, RDFS
+from rdflib import BNode, Literal, RDF, RDFS
+
 
 RDF_SEQi = "http://www.w3.org/1999/02/22-rdf-syntax-ns#_%s"
 MAX_CARD = URI("http://www.w3.org/2002/07/owl#maxCardinality")
@@ -57,7 +55,6 @@ ON_PROP = URI("http://www.w3.org/2002/07/owl#onProperty")
 ONE = Literal("1")
 
 
-
 class ThingFactory(object):
     """
     Fed a store, return a factory that can be used to instantiate
@@ -72,7 +69,7 @@ class ThingFactory(object):
         self.schema_store = schema_store or self.store
         self.alias_map = alias_map or {}
 
-    def __call__(self, ident, **props):
+    def __call__(self, ident=None, **props):
         """
         ident - either:
             a) None  (creates a new BNode)
@@ -112,7 +109,7 @@ class Thing(object):
             c) a list containing a and/or b
     """
     
-    def __init__(self, store, schema_store, alias_map, ident, props=None):
+    def __init__(self, store, schema_store, alias_map, ident=None, props=None):
         """
         store - rdflib.Graph.Graph
         schema_store - rdflib.Graph.Graph
@@ -126,12 +123,9 @@ class Thing(object):
         self._store = store
         self._schema_store = schema_store
         self._alias_map = alias_map
-        if ident is None:
-            self._id = BNode()
-        elif isinstance(ident, ID):
-            self._id = ident
-        else:
-            self._id = self._AttrToURI(ident)
+
+        self._id = self._AttrToURI(ident)
+
         if props is not None:
             for attr, obj in props.items():
                 if isinstance(obj, list):
@@ -151,20 +145,18 @@ class Thing(object):
         """
         if attr[0] == '_':
             raise AttributeError
+        elif attr.endswith('_of'):
+            pred = self._AttrToURI(attr[:-3])
+            return ResourceSet(self, pred, inverse=True)
         else:
-            if ":" in attr:
-                pred = URI(attr)
-            else:
-                try:
-                    pred = self._AttrToURI(attr)
-                except ValueError:
-                    raise AttributeError
+            pred = self._AttrToURI(attr)
+
             if self._isUniqueObject(pred):
                 try:
-                    obj = self._store.triples((self._id, pred, None)).next()[2]
+                    obj = self._store.objects(self._id, pred).next()
                 except StopIteration:
                     raise AttributeError
-                return self._rdfToPython(pred, obj)
+                return self._rdf_to_python(pred, obj)
             else:
                 return ResourceSet(self, pred)
                 
@@ -179,16 +171,12 @@ class Thing(object):
         if attr[0] == '_':
             self.__dict__[attr] = obj
         else:
-            if ":" in attr:
-                pred = URI(attr)
-            else:
-                try:                    
-                    pred = self._AttrToURI(attr)
-                except ValueError:
-                    raise AttributeError
+            pred = self._AttrToURI(attr)
+
             if self._isUniqueObject(pred):
                 self._store.remove((self._id, pred, None))
-                self._store.add((self._id, pred, self._pythonToRdf(pred, obj)))
+                obj_rdf = self._python_to_rdf(pred, obj)
+                self._store.add((self._id, pred, obj_rdf))
             elif isinstance(obj, ResourceSet) or type(obj) is type(set()):
                 ResourceSet(self, pred, obj.copy())
             else:
@@ -204,26 +192,20 @@ class Thing(object):
         if attr[0] == '_':
             del self.__dict__[attr]
         else:
-            if ":" in attr:
-                self._store.remove((self._id, URI(attr), None))
-            else:
-                try:
-                    self._store.remove((self._id, self._AttrToURI(attr), None))
-                except ValueError:
-                    raise AttributeError
+            self._store.remove((self._id, self._AttrToURI(attr), None))
 
-    def _rdfToPython(self, pred, obj):
+    def _rdf_to_python(self, pred, obj, inverse=False):
         """
         Given a RDF predicate and object, return the equivalent Python object.
         
-        pred - rdflib.URIRef.URIRef instance
-        obj - rdflib.Identifier.Identifier instance
+        pred - rdflib.URIRef instance
+        obj - rdflib.Identifier instance
 
         returns a python data representation
         """ 
-        obj_types = self._getObjectTypes(pred, obj)
+        obj_types = self._getObjectTypes(pred, obj, inverse=inverse)
         if isinstance(obj, Literal):  # typed literals
-            return self._literalToPython(obj, obj_types)
+            return obj.toPython()
         elif RDF.List in obj_types:
             return self._listToPython(obj)
         elif RDF.Seq in obj_types:
@@ -231,17 +213,17 @@ class Thing(object):
             while True:
                 counter = URI(RDF_SEQi % i)
                 try:
-                    item = self._store.triples((obj, counter, None)).next()[2]
+                    item = self._store.objects(obj, counter).next()
                 except StopIteration:
                     return l
-                l.append(self._rdfToPython(counter, item)) 
+                l.append(self._rdf_to_python(counter, item))
                 i += 1
         elif isinstance(obj, ID):
             return self.__class__(self._store, self._schema_store, self._alias_map, obj)
         else:
             raise ValueError
 
-    def _pythonToRdf(self, pred, obj, lang=None):
+    def _python_to_rdf(self, pred, obj, lang=None):
         """
         Given a Python predicate and object, return the equivalent RDF object.
         
@@ -251,70 +233,57 @@ class Thing(object):
         returns rdflib.Identifier.Identifier instance
         """
         obj_types = self._getObjectTypes(pred, obj)
+
         if RDF.List in obj_types:
             blank = BNode()
             self._pythonToList(blank, obj)   ### this actually stores things... 
             return blank
+
         elif RDF.Seq in obj_types:  ### so will this
             blank = BNode()
             i = 1
             for item in obj:
                 counter = URI(RDF_SEQi % i)
-                self._store.add((blank, counter, self._pythonToRdf(counter, item)))
+                self._store.add((blank, counter, self._python_to_rdf(counter, item)))
                 i += 1
             return blank
+
         elif isinstance(obj, self.__class__):
             if obj._store is not self._store:
                 obj.copyTo(self._store)  ### and this...
             return obj._id
-        else:
-            return self._pythonToLiteral(obj, obj_types, lang=lang)
 
-    def _literalToPython(self, obj, obj_types):
-        """
-        obj - rdflib.Literal.Literal instance
-        obj_types - iterator yielding rdflib.URIRef.URIRef instances
-        
-        returns a python literal datatype
-        """
-        for obj_type in obj_types:
-            try:
-                return SchemaToPython[str(obj_type)][0](obj)
-            except KeyError:
-                pass
-        return SchemaToPythonDefault[0](obj)
-    
-    def _pythonToLiteral(self, obj, obj_types, lang=None):
+        else:
+            return self._python_to_literal(obj, obj_types, lang=lang)
+
+    def _python_to_literal(self, obj, obj_types, lang=None):
         """
         obj - a python literal datatype
-        obj_types - iterator yielding rdflib.URIRef.URIRef instances
+        obj_types - iterator yielding rdflib.URIRef instances
         
         returns rdflib.Literal.Literal instance
         """
         for obj_type in obj_types:
-            try:
-                return Literal(SchemaToPython[str(obj_type)][1](obj), lang=lang)
-            except KeyError:
-                pass
-        return Literal(SchemaToPythonDefault[1](obj), lang=lang)
-            
+            return Literal(obj, datatype=obj_type, lang=lang)
+        return Literal(obj, lang=lang)
+
     def _listToPython(self, subj):
         """
         Given a RDF list, return the equivalent Python list.
         
-        subj - rdflib.Identifier.Identifier instance
+        subj - rdflib.Identifier instance
 
         returns list of python data representations
         """
         try:
-            first = self._store.triples((subj, RDF.first, None)).next()[2]
+            first = self._store.objects(subj, RDF.first).next()
         except StopIteration:
             return []
         try:
-            rest = self._store.triples((subj, RDF.rest, None)).next()[2]
+            rest = self._store.objects(subj, RDF.rest).next()
         except StopIteration:
             return ValueError
-        return [self._rdfToPython(RDF.first, first)] + self._listToPython(rest)  ### type first?
+        return [self._rdf_to_python(RDF.first, first)] + self._listToPython(rest)  ### type first?
 
     def _pythonToList(self, subj, members):
         """
@@ -323,7 +292,7 @@ class Thing(object):
         subj - rdflib.Identifier.Identifier instance
         members - list of python data representations
         """
-        first = self._pythonToRdf(RDF.first, members[0])
+        first = self._python_to_rdf(RDF.first, members[0])
         self._store.add((subj, RDF.first, first))
         if len(members) > 1:
             blank = BNode()
@@ -340,41 +309,39 @@ class Thing(object):
         
         returns rdflib.URIRef.URIRef instance
         """
+        if isinstance(attr, ID):
+            return attr
+
+        if attr is None:
+            return BNode()
+
+        if ':' in attr:
+            return URI(attr)
+
         if attr in self._alias_map:
             return URI(self._alias_map[attr])
         else:
             prefix, localname = attr.split("_", 1)
             return URI("".join([self._store.namespace_manager.store.namespace(prefix), localname]))
 
-    def _URIToAttr(self, uri):
-        """
-        Given a URI, return an attribute.
-        
-        uri - str that is a URI
-        
-        returns str in the form prefix_localname. Not the most efficient thing around.
-        """
-        for alias, alias_uri in self._alias_map.items():
-            if uri == alias_uri:
-                return alias
-        for ns_prefix, ns_uri in self._store.namespace_manager.namespaces():
-            if ns_uri == uri[:len(ns_uri)]:
-                return "_".join([ns_prefix, uri[len(ns_uri):]])
-        raise ValueError
 
-    def _getObjectTypes(self, pred, obj):
+    def _getObjectTypes(self, pred, obj, inverse=False):
         """
         Given a predicate and an object, return a list of the object's types.
         
-        pred - rdflib.URIRef.URIRef instance
-        obj - rdflib.Identifier.Identifier instance
+        pred - rdflib.URIRef instance
+        obj - rdflib.Identifier instance
         
-        returns list containing rdflib.Identifier.Identifier instances
+        returns list containing rdflib.Identifier instances
         """
-        obj_types = [o for (s, p, o) in self._schema_store.triples((pred, RDFS.range, None))]
+        if inverse:
+            obj_types = list(self._schema_store.objects(pred, RDFS.domain))
+        else:
+            obj_types = list(self._schema_store.objects(pred, RDFS.range))
 
         if isinstance(obj, URI):
-            obj_types += [o for (s, p, o) in self._store.triples((obj, RDF.type, None))]
+            obj_types += list(self._store.objects(obj, RDF.type))
+
         return obj_types
 
     def _isUniqueObject(self, pred):
@@ -393,7 +360,7 @@ class Thing(object):
         subj_types = [o for (_, _, o) in self._store.triples((self._id, RDF.type, None))]
         for type in subj_types:
             superclasses = [o for (s, p, o) in \
-              self._schema_store.triples((type, RDFS.subClassOf, None))]
+              self._schema_store.objects(type, RDFS.subClassOf)]
             for superclass in superclasses:
                 if (
                     (superclass, RDF.type, RESTRICTION) in self._schema_store and
@@ -409,10 +376,7 @@ class Thing(object):
         return self._id
         
     def __str__(self):
-        try:
-            return self._URIToAttr(self._id)
-        except ValueError:
-            return str(self._id)
+        return str(self._id)
                 
     def __eq__(self, other):
         if isinstance(other, self.__class__):
@@ -421,17 +385,19 @@ class Thing(object):
             return self._id == other
     
     def __ne__(self, other):
-        if self is other: return False
-        else: return True
-        
+        return not self.__eq__(other)
+
+    def __hash__(self):
+        return hash(self._id)
+
     def properties(self):
         """
         List unique properties.
         
         returns list containing self.__class__ instances
         """
-        return [self.__class__(self._store, self._schema_store, self._alias_map, p) 
-          for (s,p,o) in self._store.triples((self._id, None, None))]
+        return [self.__class__(self._store, self._schema_store, self._alias_map, p)
+                for (_, p, _) in self._store.triples((self._id, None, None))]
 
     def copyTo(self, store):
         """
@@ -445,12 +411,12 @@ class Thing(object):
                 self.__class__(self._store, self._schema_store, self._alias_map, o).copyTo(store)
         
         
-class ResourceSet:
+class ResourceSet(object):
     """
     A set interface to the object(s) of a non-unique RDF predicate. Interface is a subset
-    (har, har) of set().copy() returns a set.
+    (har, har) of set(). copy() returns a set.
     """
-    def __init__(self, subject, predicate, iterable=None):
+    def __init__(self, subject, predicate, iterable=None, inverse=False):
         """
         subject - rdflib.Identifier.Identifier instance
         predicate -  rdflib.URIRef.URIRef instance
@@ -459,86 +425,61 @@ class ResourceSet:
         self._subject = subject
         self._predicate = predicate
         self._store = subject._store
+        self._inverse = inverse
         if iterable is not None:
             for obj in iterable:
                 self.add(obj)
+
     def __len__(self):
-        return len(list(
-          self._store.triples((self._subject._id, self._predicate, None))))
-    def __contains__(self, obj):
-        if isinstance(obj, self._subject.__class__):    
+        if self._inverse:
+            return len(list(self._store.subjects(self._predicate, self._subject._id)))
+        else:
+            return len(list(self._store.objects(self._subject._id, self._predicate)))
+
+    def _obj_to_rdf(self, obj):
+        if isinstance(obj, type(self._subject)):
             obj = obj._id
+        elif self._inverse:
+            return self._subject._python_to_literal(obj, [])
         else: ### doesn't use pythonToRdf because that might store it
             obj_types = self._subject._getObjectTypes(self._predicate, obj) 
-            obj = self._subject._pythonToLiteral(obj, obj_types)
-        return (self._subject._id, self._predicate, obj) in self._store
+            obj = self._subject._python_to_literal(obj, obj_types)
+        return obj
+        
+
+    def __contains__(self, obj):
+        obj = self._obj_to_rdf(obj)
+        if self._inverse:
+            return (obj, self._predicate, self._subject._id) in self._store
+        else:
+            return (self._subject._id, self._predicate, obj) in self._store
+
     def __iter__(self):
-        for (s, p, o) in \
-          self._store.triples((self._subject._id, self._predicate, None)):
-            yield self._subject._pythonToRdf(self._predicate, o, lang=o.language)
+        if self._inverse:
+            for s in self._store.subjects(self._predicate, self._subject._id):
+                yield self._subject._rdf_to_python(self._predicate, s, inverse=True)
+        else:
+            for o in self._store.objects(self._subject._id, self._predicate):
+                yield self._subject._rdf_to_python(self._predicate, o)
+
     def copy(self):
         return set(self)
+
     def add(self, obj):
-        self._store.add((self._subject._id, self._predicate, 
-          self._subject._pythonToRdf(self._predicate, obj)))
+        rdf_obj = self._subject._python_to_rdf(self._predicate, obj)
+        if self._inverse:
+            self._store.add((rdf_obj, self._predicate, self._subject._id))
+        else:
+            self._store.add((self._subject._id, self._predicate, rdf_obj))
+
     def remove(self, obj):
         if not obj in self:
             raise KeyError
         self.discard(obj)
+
     def discard(self, obj):
-        if isinstance(obj, self._subject.__class__):
-            obj = obj._id
-        else: ### doesn't use pythonToRdf because that might store it
-            obj_types = self._subject._getObjectTypes(self._predicate, obj)
-            obj = self._subject._pythonToLiteral(obj, obj_types)
-        self._store.remove((self._subject._id, self._predicate, obj))
-    def clear(self):
-        self._store.remove((self._subject, self._predicate, None))
-
-        
-SchemaToPythonDefault = (unicode, unicode)
-SchemaToPython = {  #  (schema->python, python->schema)  Does not validate.
-    'http://www.w3.org/2001/XMLSchema#string': (unicode, unicode),
-    'http://www.w3.org/2001/XMLSchema#normalizedString': (unicode, unicode),
-    'http://www.w3.org/2001/XMLSchema#token': (unicode, unicode),
-    'http://www.w3.org/2001/XMLSchema#language': (unicode, unicode),
-    'http://www.w3.org/2001/XMLSchema#boolean': (bool, lambda i:unicode(i).lower()),
-    'http://www.w3.org/2001/XMLSchema#decimal': (float, unicode), 
-    'http://www.w3.org/2001/XMLSchema#integer': (long, unicode), 
-    'http://www.w3.org/2001/XMLSchema#nonPositiveInteger': (int, unicode),
-    'http://www.w3.org/2001/XMLSchema#long': (long, unicode),
-    'http://www.w3.org/2001/XMLSchema#nonNegativeInteger': (int, unicode),
-    'http://www.w3.org/2001/XMLSchema#negativeInteger': (int, unicode),
-    'http://www.w3.org/2001/XMLSchema#int': (int, unicode),
-    'http://www.w3.org/2001/XMLSchema#unsignedLong': (long, unicode),
-    'http://www.w3.org/2001/XMLSchema#positiveInteger': (int, unicode),
-    'http://www.w3.org/2001/XMLSchema#short': (int, unicode),
-    'http://www.w3.org/2001/XMLSchema#unsignedInt': (long, unicode),
-    'http://www.w3.org/2001/XMLSchema#byte': (int, unicode),
-    'http://www.w3.org/2001/XMLSchema#unsignedShort': (int, unicode),
-    'http://www.w3.org/2001/XMLSchema#unsignedByte': (int, unicode),
-    'http://www.w3.org/2001/XMLSchema#float': (float, unicode),
-    'http://www.w3.org/2001/XMLSchema#double': (float, unicode),  # doesn't do the whole range
-#    duration
-#    dateTime
-#    time
-#    date
-#    gYearMonth
-#    gYear
-#    gMonthDay
-#    gDay
-#    gMonth
-#    hexBinary
-    'http://www.w3.org/2001/XMLSchema#base64Binary': (base64.decodestring, lambda i:base64.encodestring(i)[:-1]),
-    'http://www.w3.org/2001/XMLSchema#anyURI': (str, str),
-}
-
-
-if __name__ == '__main__':
-    # use: "python -i laconia.py [URI for RDF file]+"
-    from rdflib.TripleStore import TripleStore
-    import sys
-    mystore = TripleStore()
-    for arg in sys.argv[1:]:
-        mystore.parse(arg)
-    thing = ThingFactory(mystore)
+        obj = self._obj_to_rdf(obj)
+        if self._inverse:
+            self._store.remove((obj, self._predicate, self._subject._id))
+        else:
+            self._store.remove((self._subject._id, self._predicate, obj))
